/* 
 * Copyright (C) 2008 Alan Ross, Tojio GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
*/


package com.tojio.wiiflashserverj;

//import regular java
import java.net.Socket;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;


public class Server extends Thread
{
	//used for log
	private final String CLASS_NAME = "[Server]";
	//small buffer. WiiFlash starts processing when pointer in its input buffer is at 57(!)
	private final int BUFFER_SIZE_OUT = 80;	// in orig WiiFlash 0.4 release this was 92
	private final int BUFFER_SIZE_IN = 4;
	
	private WiiFlashServerJ wfs;
	
    private Socket socket = null;
    
    private boolean quit = false;
    
    private DataOutputStream out;
    private DataInputStream in;
    
    private ByteBuffer outBuffer;
    private ByteBuffer inBuffer;
    
    
    /**
	 * Creates a new Server thread with given socket.
	 */
    public Server(WiiFlashServerJ wfs, Socket socket)
    {
    	this.wfs = wfs;
    	wfs.log(CLASS_NAME, "initialize");
    	
    	outBuffer = ByteBuffer.allocate(BUFFER_SIZE_OUT);
    	inBuffer = ByteBuffer.allocate(BUFFER_SIZE_IN);
    	
    	this.socket = socket;
    	
    	try
		{
		    //out = new DataOutputStream(new BufferedOutputStream(socket.getOutputStream(),BUFFER_SIZE_OUT));
		    out = new DataOutputStream(socket.getOutputStream());
		    in = new DataInputStream( socket.getInputStream() );
		}
    	catch (IOException e) { e.printStackTrace(); }
    }
    
    /**
     * Run the Server, This is send and listen in infinity loop on its port until quit.
     */
    public void run()
    {
    	wfs.log(CLASS_NAME, "start running");
    	
    	try
		{
			// make sending faster
    		socket.setTcpNoDelay(true);
			
		    while (!quit)
			{
		    	//slow down a little bit, input just comes seldom, and is not timing relevant
				Thread.sleep(50);
				
				//get input from WiiFlash client
				inBuffer.clear();
				inBuffer.position(0);
				inBuffer.put((byte) in.readByte());
				inBuffer.put((byte) in.readByte());
				inBuffer.put((byte) in.readByte());
				inBuffer.put((byte) in.readByte());
				inBuffer.position(0);
				
				//process input
				if(inBuffer.array() != null)
				{
					//get ID of Wiiremote
					if(inBuffer.array()[0] >= 0 && inBuffer.array()[0] <= 3)
					{
						//handle rumble request
						if(inBuffer.array()[1] == 0x72)
						{
							if(inBuffer.array()[2] == 0x31)
							{ wfs.startVibrating(inBuffer.array()[0]); }
							if(inBuffer.array()[2] == 0x30)
							{ wfs.stopVibrating(inBuffer.array()[0]); }
						}
						//handle LED request
						if(inBuffer.array()[1] == 0x6c)
						{ wfs.setLEDLights(inBuffer.array()[0], inBuffer.array()[2]); }
						//handle mouse control request
						if(inBuffer.array()[1] == 0x76)
						{
							if(inBuffer.array()[2] == 0x31)
							{
								wfs.log(CLASS_NAME, "client request: Wii["+inBuffer.array()[0]+"] start mouse control");
							}
							if(inBuffer.array()[2] == 0x30)
							{
								wfs.log(CLASS_NAME, "client request: Wii["+inBuffer.array()[0]+"] stop mouse control");
							}
						}
					}
				}
			}
		    
		    wfs.log(CLASS_NAME, "has quit");
	    	out.close();
    		in.close();
    		socket.close();
		}
		catch (Exception e) { e.printStackTrace(); }
    }
    
    /**
     * Return true if Server is running, false if not.
     */
    public boolean isRunning()
    {	
    	return !quit;
    }
    
    
    /**
     * Quit running server.
     */
    public void quit()
    {
    	wfs.log(CLASS_NAME, "quitting");
    	quit = true;
    }
    
    
    
    public void sendRemoteData(WiiMoteData data)
	{
    	try
    	{
	    	outBuffer.clear();							// clear byte buffer
			outBuffer.position(0);						// set pointer to 0
			
			outBuffer.put((byte) data.id);				// wiimote id (0 <= id <= 3)
			outBuffer.put((byte) data.battery);			// wiimote battery level
			
			outBuffer.putShort((short) data.mButtons);	// mote button ids & state (short)
			outBuffer.putFloat((float) data.mAccelX);	// mote x acceleration
			outBuffer.putFloat((float) data.mAccelY);	// mote y acceleration
			outBuffer.putFloat((float) data.mAccelZ);	// mote z acceleration
			
			outBuffer.put((byte) data.hasExtension);	// has extension (0=none, 1=nunchuck, 2=classic)
			
			if(data.hasExtension == 0)					// no extension
			{
				outBuffer.position(38);					//
			}
			if(data.hasExtension == 1)					// nunchuck <- CAREFULL: NO EXTENSION HAS BEEN TASTED YET, THIS IS BLIND GUESSING
			{
				outBuffer.put((byte) data.nButtons);		// button ids & state (byte)
				outBuffer.putFloat((float) data.nStickX);	// stick X
				outBuffer.putFloat((float) data.nStickY);	// stick Y
				outBuffer.putFloat((float) data.nAccelX);	// nunchuck accel x
				outBuffer.putFloat((float) data.nAccelY);	// nunchuck accel y
				outBuffer.putFloat((float) data.nAccelZ);	// nunchuck accel z
			}
			if(data.hasExtension == 2)					// classic controller <- CAREFULL: NO EXTENSION HAS BEEN TASTED YET, THIS IS BLIND GUESSING
			{
				outBuffer.putShort((short) data.cButtons);	// button ids & state (short)
				outBuffer.putFloat((float) data.cStickLX);	// stickXLeft
				outBuffer.putFloat((float) data.cStickLY);	// stickYLeft
				outBuffer.putFloat((float) data.cStickRX);	// stickXRight
				outBuffer.putFloat((float) data.cStickLY);	// stickYRight
			}
			
			if( data.irP[0].x != -10 ){ outBuffer.put((byte) 1); }	// point1 is active
			else{ outBuffer.put((byte) 0); }
			outBuffer.putFloat((float) data.irP[0].x);		// ir x1
			outBuffer.putFloat((float) data.irP[0].y);		// ir y1
			
			if( data.irP[1].x != -10 ){ outBuffer.put((byte) 1); }	// point2 is active
			else{ outBuffer.put((byte) 0); }
			outBuffer.putFloat((float) data.irP[1].x);		// ir x2
			outBuffer.putFloat((float) data.irP[1].y);		// ir y2
			
			if( data.irP[2].x != -10 ){ outBuffer.put((byte) 1); }	// point3 is active (WiiFlash can't process this yet)
			else{ outBuffer.put((byte) 0); }
			outBuffer.putFloat((float) data.irP[2].x);		// ir x3
			outBuffer.putFloat((float) data.irP[2].y);		// ir y3
			
			if( data.irP[3].x != -10 ){ outBuffer.put((byte) 1); }	// point4 is active (WiiFlash can't process this yet)
			else{ outBuffer.put((byte) 0); }
			outBuffer.putFloat((float) data.irP[3].x);		// ir x4
			outBuffer.putFloat((float) data.irP[3].y);		// ir y4
			
			outBuffer.put((byte) data.irS[0]);				// ir size of p1
			outBuffer.put((byte) data.irS[1]);				// ir size of p2
			outBuffer.put((byte) data.irS[2]);				// ir size of p3
			outBuffer.put((byte) data.irS[3]);				// ir size of p4
			
			
			
			//outBuffer.put((byte) 0);						//fill up the buffer
			//outBuffer.put((byte) 0);						//fill up the buffer
			
			//write output to socket / WiiFlash
			if(outBuffer.array() != null)
			{
				out.write(outBuffer.array());
				out.flush();
				//wfs.log(CLASS_NAME, "send remote data (vars)");
			}
		}
		catch(Exception e){}
	}
    
    
    /*
    // Deprecated
    public void sendRemoteData(int id, int mbuttons, float[] mAccel, java.awt.geom.Point2D[] mIR, int hasExtension)
	{
    	try
    	{
	    	outBuffer.clear();								// clear byte buffer
			outBuffer.position(0);							// set pointer to 0
			
			outBuffer.put((byte) id);						// wiimote id (0 <= id <= 3)
			outBuffer.putShort((short) mbuttons);			// button ids & state (short)
			outBuffer.putFloat((float) mAccel[0]);			// x acceleration
			outBuffer.putFloat((float) mAccel[1]);			// y acceleration
			outBuffer.putFloat((float) mAccel[2]);			// z acceleration
			outBuffer.put((byte) hasExtension);				// has extension (0=none, 1=nunchuck, 2=classic)
			
			if(hasExtension == 0)							// no extension
			{
				outBuffer.put((byte) 0);					// (fake input, just fill space)
				outBuffer.putFloat(0);
				outBuffer.putFloat(0);
				outBuffer.putFloat(0);
				outBuffer.putFloat(0);
				outBuffer.putFloat(0);
			}
			if(hasExtension == 1)							// nunchuck
			{
				outBuffer.put((byte) 1024);					// (fake input) button ids & state (byte)
				outBuffer.putFloat((float)1024);			// (fake input) stick X
				outBuffer.putFloat((float)1024);			// (fake input) stick Y
				outBuffer.putFloat((float)1024);			// (fake input) nun x
				outBuffer.putFloat((float)1024);			// (fake input) nun y
				outBuffer.putFloat((float)1024);			// (fake input) nun z
			}
			if(hasExtension == 2)							// classic controller
			{
				outBuffer.putShort((short)2048);			// (fake input) button ids & state (short)
				outBuffer.putFloat((float)2048);			// (fake input) stickXLeft
				outBuffer.putFloat((float)2048);			// (fake input) stickYLeft
				outBuffer.putFloat((float)2048);			// (fake input) stickXRight
				outBuffer.putFloat((float)2048);			// (fake input) stickYRight
			}
			
			if( mIR[0].x != -10 ){ outBuffer.put((byte) 1); }	// point1 is active
			else{ outBuffer.put((byte) 0); }
			outBuffer.putFloat((float) mIR[0].x);	// ir x1
			outBuffer.putFloat((float) mIR[0].y);	// ir y1
			
			if( mIR[1].x != -10 ){ outBuffer.put((byte) 1); }	// point2 is active
			else{ outBuffer.put((byte) 0); }
			outBuffer.putFloat((float) mIR[1].x);	// ir x2
			outBuffer.putFloat((float) mIR[1].y);	// ir y2
			
			outBuffer.put((byte) 0);							//fill up the buffer
			outBuffer.put((byte) 0);							//fill up the buffer
			
			//write output to socket / WiiFlash
			if(outBuffer.array() != null)
			{
				out.write(outBuffer.array());
				out.flush();
				//wfs.log(CLASS_NAME, "send remote data (vars)");
			}
		}
		catch(Exception e){}
	}
	
    // Deprecated
    public void sendRemoteData(ByteBuffer data )
    {
    	try
    	{
    		if(data.array() != null)
			{
    			outBuffer.limit(BUFFER_SIZE_OUT);
				out.write(data.array());
				out.flush();
				//wfs.log(CLASS_NAME, "send remote data (bytebuffer)");
			}
    	}
    	catch(Exception e){}
    }
    */
    
    
}