/* 
 * Copyright (C) 2008 Alan Ross, Tojio GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * Note: WiiMoteAdapter.java has made extensive use of the demo-function-code 
 * (WRLImpl.java) provided by the WiiRemoteJ library by Michael Diamond.
*/


package com.tojio.wiiflashserverj;

//import WiiremoteJ
import wiiremotej.*;
import wiiremotej.event.*;


public class WiiMoteAdapter extends WiiRemoteAdapter
{
	//used for log
	private final String CLASS_NAME = "[WiiMoteAdapter]";
	
	private WiiFlashServerJ wfs;
	private WiiRemote remote;
	private WiiMoteData data;
	
	
    
    /**
     * Creates a new WiiMoteAdapter object
     */
    public WiiMoteAdapter(WiiFlashServerJ wfs, WiiRemote remote)
    {
    	this.wfs = wfs;
        this.remote = remote;
        
        data = new WiiMoteData();

        data.id = wfs.registerRemote(this);
        
        if(data.id ==-1){ disconnect(); } //if 4 WiiRemotes are all ready connected. maximum is reached...
        else{ init(); }
    }
    
    /**
     * 
     */
    public void init()
    {
    	wfs.log(CLASS_NAME + "["+data.id+"]", "starting up");
        
        try
        {
        	wfs.log(CLASS_NAME + "["+data.id+"]", "connect");
        	//remote.enableContinuous();
        	remote.addWiiRemoteListener(this);
            remote.setAccelerometerEnabled(true);
            remote.setSpeakerEnabled(true);
            remote.setIRSensorEnabled(true, WRIREvent.EXTENDED);
            remote.setLEDIlluminated(data.id, true);
            
            remote.requestStatus();
            
            remote.getButtonMaps().add(new ButtonMap(WRButtonEvent.HOME, ButtonMap.NUNCHUK, 
            		WRNunchukExtensionEvent.C, new int[]{java.awt.event.KeyEvent.VK_CONTROL}, 
            		java.awt.event.InputEvent.BUTTON1_MASK, 0, -1));
        }
        catch(Exception e)
        { 
        	wfs.log(CLASS_NAME + "["+data.id+"]", "Error " + e.getMessage());
        	//should we disconnect() if there has been an error??
        }
    }
    
    
    /**
     * Get Bluetooth address of this WiiRemote
     */
    public String getBluetoothAddress()
    {
    	String s = "";
    	try{ s = remote.getBluetoothAddress(); }
    	catch(Exception e){}
    	return s;
    }
    
    
    /**
     * Get the ID of this WiiRemote
     */
    public int getID()
    {
    	return data.id;
    }
    
    
    /**
     * Set the LED lights of this WiiRemote
     * @param lightID 0=Light1, 1=Light2 ...
     */
    public void setLEDLights(int lightID)
    {
    	try
    	{
    		if(lightID == 0){ remote.setLEDLights(new boolean[] {true,false,false,false}); }
    		if(lightID == 1){ remote.setLEDLights(new boolean[] {false,true,false,false}); }
    		if(lightID == 2){ remote.setLEDLights(new boolean[] {false,false,true,false}); }
    		if(lightID == 3){ remote.setLEDLights(new boolean[] {false,false,false,true}); }
    	}
    	catch(Exception e){}
    }
    
    
    /**
     * Start vibrating this WiiRemote
     */
    public void startVibrating()
    {
    	try{ remote.startVibrating(); }
    	catch(Exception e){}
    }
    
    
    /**
     * Stop vibrating this WiiRemote
     */
    public void stopVibrating()
    {
    	try{ remote.stopVibrating(); }
    	catch(Exception e){}
    }
    
    
    /**
	 * Disconnect this WiiRemote
	 */
	public void disconnect()
	{
		wfs.log(CLASS_NAME + "["+data.id+"]", "disconnect");
		if(remote != null || isConnected())
		{ remote.disconnect(); }
	}
	
	
	/**
	 * Return true if WiiRemote is connected, false if not
	 */
	public boolean isConnected()
	{
		if(remote != null) return remote.isConnected();
		else return false;
	}
    
	
    /**
	 * Fires whenever input is received on the regular input reports (0x30...0x3f).
	 * 
	 * [NOTE] Buttons: WiiFlash uses bit shifting and the logic AND operator to find the buttons that 
	 * were pressed.
	 * if the button e.g. LEFT is pressed then 0000.0000.0000.0000.0000.0000.0001.0000 is sent = 32
	 * if the button e.g. RIGHT is pressed then 0000.0000.0000.0000.0000.0000.0010.0000 is sent = 64
	 * this way you buttons id and state are encoded into one single number. It also is possible
	 * to encode states of several buttons into that number:
	 * If button LEFT & RIGHT are pressed then 0000.0000.0000.0000.0000.0000.0011.0000 is sent = 96
	 * 
	 * EXTENSION DATA IS NOT IMPLEMENTED / PROCESSED YET
	 */
	public void	combinedInputReceived(WRCombinedEvent evt) 
    {
		try
		{
			//get remote button input
			data.mButtons = 0;
			if (evt.getButtonEvent().isPressed(WRButtonEvent.ONE)){ data.mButtons += data.ONE; }
			if (evt.getButtonEvent().isPressed(WRButtonEvent.TWO)){ data.mButtons += data.TWO; }
			if (evt.getButtonEvent().isPressed(WRButtonEvent.A)){ data.mButtons += data.A; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.B)){ data.mButtons += data.B; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.PLUS)){ data.mButtons += data.PLUS; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.MINUS)){ data.mButtons += data.MINUS; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.HOME)){ data.mButtons += data.HOME; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.UP)){ data.mButtons += data.UP; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.DOWN)){ data.mButtons += data.DOWN; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.RIGHT)){ data.mButtons += data.RIGHT; }
	        if (evt.getButtonEvent().isPressed(WRButtonEvent.LEFT)){ data.mButtons += data.LEFT; }
	        
	        //get remote acceleration input
	        data.mAccelX = (float) evt.getAccelerationEvent().getXAcceleration();
	        data.mAccelY = (float) evt.getAccelerationEvent().getYAcceleration(); 
	        data.mAccelZ = (float) evt.getAccelerationEvent().getZAcceleration();
			
			//get remote IR input
	        data.irP[0].setLocation(-10,-10);
	        data.irP[1].setLocation(-10,-10);
	        data.irP[2].setLocation(-10,-10);
	        data.irP[3].setLocation(-10,-10);
	        data.irS[0] = 0;
	        data.irS[1] = 0;
	        data.irS[2] = 0;
	        data.irS[3] = 0;
			
			int lightCounter = 0;
			for (IRLight light : evt.getIREvent().getIRLights())
	        {
	            if (light != null)
	            {
	            	data.irP[lightCounter].setLocation((float)light.getX(),-(float)light.getY());
	            	//WiiFlash wants point size 0-15;
	            	data.irS[lightCounter] = (int) ((light.getSize()*100)/6.7);
	            	//wfs.log("P"+lightCounter, "s:"+ data.irS[lightCounter] +" x:"+ data.irP[lightCounter].x + " y:" + data.irP[lightCounter].y);
	            	lightCounter++;
	            }
	        }
		
			//get extensions input
			data.hasExtension = 0;
			//Nunchuck (NOT IMPLEMENTED YET, JUST GUESSING...)
			if (evt.getExtensionEvent() instanceof WRNunchukExtensionEvent)
	        { 
				data.hasExtension = 1; 
				WRNunchukExtensionEvent NEvt = (WRNunchukExtensionEvent)evt.getExtensionEvent();
		        
				data.nButtons = 0;
				if (NEvt.isPressed(WRNunchukExtensionEvent.C)){ data.nButtons += data.NC; }
				if (NEvt.isPressed(WRNunchukExtensionEvent.Z)){ data.nButtons += data.NZ; }
				
				data.nStickX = (float) NEvt.getAnalogStickData().getX();
				data.nStickY = (float) NEvt.getAnalogStickData().getY();
				
				data.nAccelX = (float) NEvt.getAcceleration().getXAcceleration();
				data.nAccelY = (float) NEvt.getAcceleration().getYAcceleration();
				data.nAccelZ = (float) NEvt.getAcceleration().getZAcceleration();				
			}
			//Classic controller (NOT IMPLEMENTED YET, JUST GUESSING...)
	        else if (evt.getExtensionEvent() instanceof WRClassicControllerExtensionEvent)
	        { 
	        	data.hasExtension = 2;
				WRClassicControllerExtensionEvent CEvt = (WRClassicControllerExtensionEvent)evt.getExtensionEvent();
				
				data.cButtons = 0;
				if (CEvt.isPressed(WRClassicControllerExtensionEvent.A)){ data.cButtons += data.A; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.B)){ data.cButtons += data.B; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.DPAD_LEFT)){ data.cButtons += data.LEFT; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.DPAD_RIGHT)){ data.cButtons += data.RIGHT; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.DPAD_UP)){ data.cButtons += data.UP; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.DPAD_DOWN)){ data.cButtons += data.DOWN; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.PLUS)){ data.cButtons += data.PLUS; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.MINUS)){ data.cButtons += data.MINUS; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.HOME)){ data.cButtons += data.HOME; }
		        
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.Y)){ data.cButtons += data.Y; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.X)){ data.cButtons += data.X; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.LEFT_Z)){ data.cButtons += data.ZL; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.RIGHT_Z)){ data.cButtons += data.ZR; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.LEFT_TRIGGER)){ data.cButtons += data.L; }
		        if (CEvt.isPressed(WRClassicControllerExtensionEvent.RIGHT_TRIGGER)){ data.cButtons += data.R; }
		        
		        data.cStickLX = (float) CEvt.getLeftAnalogStickData().getX();
		        data.cStickLY = (float) CEvt.getLeftAnalogStickData().getY();
		        data.cStickRX= (float) CEvt.getRightAnalogStickData().getX();
		        data.cStickRY= (float) CEvt.getRightAnalogStickData().getY();
		        
		        //CEvt.getLeftTrigger(); ???
		        //CEvt.getRightTrigger(); ???
			}
			//Guitar (NOT SUPPORTED BY WIIFLASH)
	        else if (evt.getExtensionEvent() instanceof WRGuitarExtensionEvent)
	        { data.hasExtension = 3; wfs.log(CLASS_NAME + "["+data.id+"]", "guitar extension"); }
			
			
			//wfs.log(CLASS_NAME + "["+data.id+"]", "sending data");
			wfs.sendRemoteData(data);
		}
		catch(Exception e){}
	}
	
	
	
	
	/**
	 * Fires whenever data is received on the status report.
	 */
	public void	statusReported(WRStatusEvent evt)
	{
		// no constant update, just at beginning...
		// the multiplication by 100 is used to provide the correct % for WiiFlash processing, which is: battery value/0xC8 
		data.battery = (float) (evt.getBatteryLevel()*100);
		wfs.log(CLASS_NAME + "["+data.id+"]", "status | Battery level: " + (data.battery/2)+ "%");
		//wfs.log(CLASS_NAME + "["+data.id+"]", "status | Battery level: " + (double)evt.getBatteryLevel()/2+ "%");
		wfs.log(CLASS_NAME + "["+data.id+"]", "status | Continuous: " + evt.isContinuousEnabled());
		wfs.log(CLASS_NAME + "["+data.id+"]", "status | Remote continuous: " + remote.isContinuousEnabled());
	}
	
	
	/**
	 * Fires when the WiiRemote disconnects.
	 */
	public void	disconnected() 
    {
		 wfs.log(CLASS_NAME + "["+data.id+"]", "disconnected");
	}

	
	/**
	 * Fires when an extension is connected and properly setup.
	 */      
	public void	extensionConnected(WiiRemoteExtension extension) 
    {
		wfs.log(CLASS_NAME + "["+data.id+"]", "Extension connected");
	}
	
	
	/**
	 * Fires when an extension disconnects.
	 */
	public void	extensionDisconnected(WiiRemoteExtension extension) 
    {
		wfs.log(CLASS_NAME + "["+data.id+"]", "Extension disconnected");
	}
	
	
	
	/**
	 * Fires when an extension is partially inserted.
	 */
	public void	extensionPartiallyInserted() 
	{
		wfs.log(CLASS_NAME + "["+data.id+"]", "Put in Extension properly");
	}
	
	
	/**
	 * Fires when an extension of an unknown type connects.
	 */
	public void	extensionUnknown() 
    {
		wfs.log(CLASS_NAME + "["+data.id+"]", "Extension unknown");
	}
	
	/**
	 * Fires whenever data is received from the accelerometer. 
	 */
	public void accelerationInputReceived(WRAccelerationEvent evt)
	{
		//wfs.log(CLASS_NAME + "["+id+"]", "Acceleration Input received\n");
	}


	/**
	 * Fires whenever data is received from the button.
	 */
	public void	buttonInputReceived(WRButtonEvent evt) 
	{
		//wfs.log(CLASS_NAME + "["+id+"]", "Button Input received\n");
	}
	
	
	/**
	 * Fires whenever data is received from IR sensor.
	 */
	public void	IRInputReceived(WRIREvent evt) 
    {
		//wfs.log(CLASS_NAME + "["+id+"]", "IR Input received\n");
	}
	
	
	/**
	 * Fires whenever there's input from the extension port.
	 */
	public void	extensionInputReceived(WRExtensionEvent evt) 
    {
		//wfs.log(CLASS_NAME + "["+id+"]", "Extension Input received");
	}
    
}